/*
 * Copyright (C) 2012 Canonical, Inc
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 *
 * Authors:
 *      Alberto Mardegan <alberto.mardegan@canonical.com>
 */

public class OnlineAccounts.Plugins.LaunchpadAccount : OnlineAccounts.Account {
    string method = "launchpad";
    public Ag.AuthData auth_data;
    public Signon.IdentityInfo info;
    private Signon.Identity identity;
    bool is_new = false;

    public LaunchpadAccount (Ag.Account account, bool is_new = false) {
        this.account = account;
        this.is_new = is_new;
        var account_service = new Ag.AccountService (account, null);
        auth_data = account_service.get_auth_data ();
        session_data = auth_data.get_login_parameters (null);
        if (is_new) {
            setup_authentification ();
        }
    }

    public override void setup_authentification () {
        info = new Signon.IdentityInfo ();
        info.set_caption (account.get_provider_name ());
        info.set_identity_type (Signon.IdentityType.APP);
        info.set_secret ("", true);
        info.set_method (method, {"launchpad", null});
        info.access_control_list_append (new Signon.SecurityContext.from_values ("%s/bin/switchboard".printf (Build.CMAKE_INSTALL_PREFIX), "*"));
        var allowed_realms = session_data.lookup_value ("AllowedRealms", null).dup_strv ();
        info.set_realms (allowed_realms);
        identity = new Signon.Identity ();
        identity.store_credentials_with_info (info, store_credentials_callback);
    }

    public async void authenticate (Signon.Identity self, uint32 id) {
        GLib.Variant? v_id = new GLib.Variant.uint32 (id);
        account.set_variant (gsignon_id, v_id);

        try {
            var session = identity.create_session (method);
            session_result = yield session.process_async (session_data, method, null);
            var consumer_key = session_data.lookup_value ("ConsumerKey", null).dup_string ();
            var token = session_result.lookup_value ("AccessToken", null).dup_string ();
            var token_secret = session_result.lookup_value ("TokenSecret", null).dup_string ();
            var now = new DateTime.now_utc ();
            var now_unix = now.to_unix ();
            var rand = new Rand.with_seed ((uint32) now_unix);

            var builder = new StringBuilder ("OAuth ");
            Soup.header_g_string_append_param_quoted (builder, "realm", "https://api.launchpad.net/");
            builder.append (",");
            Soup.header_g_string_append_param_quoted (builder, "oauth_consumer_key", consumer_key);
            builder.append (",");
            Soup.header_g_string_append_param_quoted (builder, "oauth_token", token);
            builder.append (",");
            Soup.header_g_string_append_param_quoted (builder, "oauth_signature_method", "PLAINTEXT");
            builder.append (",");
            Soup.header_g_string_append_param_quoted (builder, "oauth_signature", "%%26%s".printf (token_secret));
            builder.append (",");
            Soup.header_g_string_append_param_quoted (builder, "oauth_timestamp", "%lld".printf (now_unix));
            builder.append (",");
            Soup.header_g_string_append_param_quoted (builder, "oauth_nonce", "%u".printf (rand.next_int ()));
            builder.append (",");
            Soup.header_g_string_append_param_quoted (builder, "oauth_version", "1.0");
            var message = new Soup.Message ("GET", "https://api.launchpad.net/devel/people/+me");
            message.request_headers.append ("Authorization", builder.str);
            var soup_session = new Soup.Session ();
            soup_session.send_message (message);

            try {
                var parser = new Json.Parser ();
                var body = message.response_body;
                parser.load_from_data ((string) body.data, (ssize_t) body.length);
                var name = parser.get_root ().get_object ().get_string_member ("name");
                account.set_display_name (name);
            } catch (Error e) {
                critical (e.message);
            }

            info.set_secret (token, true);
            identity.query_info (query_info_callback);
        } catch (Error e) {
            critical (e.message);
        }

        yield;
    }

    // Callbacks
    [CCode (instance_pos = -1)]
    public void store_credentials_callback (Signon.Identity self, uint32 id, GLib.Error error) {
        if (error != null) {
            critical (error.message);
            return;
        }

        authenticate.begin (self, id);
    }

    [CCode (instance_pos = -1)]
    public void query_info_callback (Signon.Identity self, Signon.IdentityInfo info, GLib.Error error) {
        if (error != null) {
            critical (error.message);
            return;
        }

        account.set_enabled (true);
        account.store_async.begin (null);
        if (is_new == true) {
            AccountsManager.get_default ().add_account (this);
            is_new = false;
        }
    }
}
