// -*- Mode: vala; indent-tabs-mode: nil; tab-width: 4 -*-
/*-
 * Copyright (c) 2014 Pantheon Developers (http://launchpad.net/online-accounts-plug)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authored by: Corentin Noël <tintou@mailoo.org>
 */

public class GSignond.LaunchpadPlugin : GLib.Object, GSignond.Plugin {
    public string type { owned get { return "launchpad"; } }
    public string[] mechanisms { owned get { return {"launchpad"}; } }
    private const string RequestEndpoint = "https://launchpad.net/+request-token";
    private const string AuthorizationEndpoint = "https://launchpad.net/+authorize-token";
    private const string TokenEndpoint = "https://launchpad.net/+access-token";
    private const string CallbackURL = "https://elementary.io";
    private GSignond.SessionData oauth_request;
    private GSignond.Dictionary token_cache;
    private Soup.Session soup_session;

    public void cancel () {
        var signond_error = new GSignond.Error.SESSION_CANCELED ("Session canceled");
        error (signond_error);
    }

    public void request_initial (GSignond.SessionData session_data, GSignond.Dictionary token_cache, string mechanism) {
        if (soup_session == null) {
            soup_session = new Soup.Session ();
            soup_session.add_feature_by_type (typeof (Soup.ProxyResolverDefault));
            soup_session.ssl_use_system_ca_file = true;
            var logger = new Soup.Logger (Soup.LoggerLogLevel.BODY, -1);
            logger.set_printer (LoggerPrinter);
            soup_session.add_feature (logger);
        }

        reset ();

        string consumer_key = session_data.get_string ("ConsumerKey");
        if (consumer_key == null) {
            var signond_error = new GSignond.Error.MISSING_DATA ("Client did not supply ConsumerKey");
            error (signond_error);
            return;
        }

        if (token_cache == null) {
            var signond_error = new GSignond.Error.MISSING_DATA ("Client did not supply token cache");
            error (signond_error);
            return;
        }

        var token_variant = token_cache.get (consumer_key);
        if (token_variant != null) {
            var token = new GSignond.Dictionary.from_variant (token_variant);
            var response = respond_with_stored_token (token);
            if (response != null) {
                response_final (response);
                return;
            }
        }

        this.oauth_request = session_data;
        this.token_cache = token_cache;

        request_temporary_token ();
    }

    public void request (GSignond.SessionData session_data) {
        var err = new GSignond.Error.WRONG_STATE ("Launchpad plugin doesn't support request");
        error (err); 
    }

    public void user_action_finished (GSignond.SignonuiData ui_data) {
        GSignond.SignonuiError query_error;
        if (ui_data.get_query_error (out query_error) == false) {
            var err = new GSignond.Error.USER_INTERACTION ("user_action_finished did not return an error value");
            error (err);
            return;
        }

        if (query_error == GSignond.SignonuiError.CANCELED) {
            var err = new GSignond.Error.SESSION_CANCELED ("Session canceled");
            error (err);
            return;
        } else if (query_error != GSignond.SignonuiError.NONE) {
            var err = new GSignond.Error.USER_INTERACTION ("user_action_finished error: %d", query_error);
            error (err);
            return;
        }

        var response_url = ui_data.get_url_response ();
        if (response_url == null || !response_url.has_prefix (CallbackURL)) {
            var err = new GSignond.Error.NOT_AUTHORIZED ("Callback URI and URI supplied by UI don't match");
            error (err);
            return;
        }

        oauth_request.remove ("Callback");

        request_access_token ();
    }

    public void refresh (GSignond.SignonuiData session_data) {
        var err = new GSignond.Error.WRONG_STATE ("Launchpad plugin doesn't support refresh");
        error (err); 
    }

    private void request_temporary_token () {
        var endpoint_uri = new Soup.URI (RequestEndpoint);
        var authorization_header = make_authorization_header (oauth_request, endpoint_uri);
        if (authorization_header != null) {
            var msg = new Soup.Message.from_uri ("POST", endpoint_uri);
            msg.set_request ("application/x-www-form-urlencoded", Soup.MemoryUse.STATIC, {});
            msg.request_headers.append ("Authorization", authorization_header);
            soup_session.queue_message (msg, temporary_token_callback);
        }
    }

    private void request_access_token () {
        var endpoint_uri = new Soup.URI (TokenEndpoint);

        var header = new StringBuilder ("");
        var oauth_consumer_key = oauth_request.get_string ("ConsumerKey");
        if (oauth_consumer_key == null) {
            var signond_error = new GSignond.Error.MISSING_DATA ("Client did not supply ConsumerKey");
            error (signond_error);
            return;
        }

        Soup.header_g_string_append_param (header, "oauth_consumer_key", oauth_consumer_key);
        header.append ("&");

        var oauth_temp_token = oauth_request.get_string ("_OauthTemporaryToken");
        if (oauth_temp_token != null) {
            Soup.header_g_string_append_param (header, "oauth_token", oauth_temp_token);
            header.append ("&");
        }

        var oauth_temp_token_secret = oauth_request.get_string ("_OauthTemporaryTokenSecret");
        Soup.header_g_string_append_param (header, "oauth_signature", "%26"+oauth_temp_token_secret);
        header.append ("&");

        Soup.header_g_string_append_param (header, "oauth_signature_method", "PLAINTEXT");
        var authorization_header = header.str;
        if (authorization_header != null) {
            var msg = new Soup.Message.from_uri ("POST", endpoint_uri);
            msg.set_request ("application/x-www-form-urlencoded", Soup.MemoryUse.COPY, authorization_header.data);
            soup_session.queue_message (msg, access_token_callback);
        }
    }

    [CCode (instance_pos = -1)]
    private void temporary_token_callback (Soup.Session session, Soup.Message msg) {
        if (msg.status_code != Soup.Status.OK) {
            reset ();
            var signond_error = new GSignond.Error.NOT_AUTHORIZED ("Temporary token endpoint returned an error: %u %s", msg.status_code, msg.reason_phrase);
            error (signond_error);
            return;
        }

        var response = Soup.Form.decode ((string) msg.response_body.flatten ().data);
        var token = response.lookup ("oauth_token");
        var token_secret = response.lookup ("oauth_token_secret");
        if (token == null || token_secret == null) {
            reset ();
            var signond_error = new GSignond.Error.NOT_AUTHORIZED ("Temporary token endpoint returned an invalid response");
            error (signond_error);
            return;
        }

        var authorization_url = new Soup.URI (AuthorizationEndpoint);
        var query = new HashTable<string, string> (str_hash, str_equal);
        var authorization_query_s = authorization_url.get_query ();
        if (authorization_query_s != null) {
            Soup.Form.decode (authorization_query_s).foreach ((key, value) => {
                query.insert (key, value);
            });
        }

        query.insert ("oauth_token", token);
        query.insert ("oauth_callback", CallbackURL);
        authorization_url.set_query_from_form (query);

        var open_url = authorization_url.to_string (false);
        oauth_request.set_string ("_OauthTemporaryToken", token);
        oauth_request.set_string ("_OauthTemporaryTokenSecret", token_secret);

        var ui_request = (GSignond.SignonuiData) new GSignond.Dictionary ();
        ui_request.set_open_url (open_url);
        ui_request.set_final_url (CallbackURL);
        user_action_required (ui_request);
    }

    [CCode (instance_pos = -1)]
    private void access_token_callback (Soup.Session session, Soup.Message msg) {
        if (msg.status_code != Soup.Status.OK) {
            reset ();
            var signond_error = new GSignond.Error.NOT_AUTHORIZED ("Access token endpoint returned an error: %u %s", msg.status_code, msg.reason_phrase);
            error (signond_error);
            return;
        }

        var response = Soup.Form.decode ((string) msg.response_body.flatten ().data);
        var token = response.lookup ("oauth_token");
        var token_secret = response.lookup ("oauth_token_secret");
        if (token == null || token_secret == null) {
            reset ();
            var signond_error = new GSignond.Error.NOT_AUTHORIZED ("Access token endpoint returned an invalid response");
            error (signond_error);
            return;
        }

        var token_dict = new GSignond.Dictionary ();
        token_dict.set_string ("AccessToken", token);
        token_dict.set_string ("TokenSecret", token_secret);

        var parameters_dict = new GSignond.Dictionary ();
        response.foreach ((key, value) => {
            if (key != "oauth_token" && key != "oauth_token_secret") {
                parameters_dict.set_string (key, value);
            }
        });

        token_dict.set ("TokenParameters", parameters_dict.to_variant ());

        var client_id = oauth_request.get_string ("ConsumerKey");
        token_cache.set (client_id, token_dict.to_variant ());

        store (token_cache);
        response_final (token_dict);
    }

    private string? make_authorization_header(GSignond.SessionData session_data, Soup.URI uri) {
        var header = new StringBuilder ("OAuth ");

        var oauth_consumer_key = session_data.get_string ("ConsumerKey");
        if (oauth_consumer_key == null) {
            var signond_error = new GSignond.Error.MISSING_DATA ("Client did not supply ConsumerKey");
            error (signond_error);
            return null;
        }

        Soup.header_g_string_append_param_quoted (header, "oauth_consumer_key", oauth_consumer_key);
        header.append (", ");

        var oauth_temp_token = session_data.get_string ("_OauthTemporaryToken");
        if (oauth_temp_token != null) {
            Soup.header_g_string_append_param_quoted (header, "oauth_token", oauth_temp_token);
            header.append (", ");
        }

        Soup.header_g_string_append_param_quoted (header, "oauth_callback", CallbackURL);
        header.append (", ");

        Soup.header_g_string_append_param_quoted (header, "oauth_signature", "%26");
        header.append (", ");

        Soup.header_g_string_append_param_quoted (header, "oauth_signature_method", "PLAINTEXT");
        header.append (", ");

        Soup.header_g_string_append_param_quoted (header, "oauth_version", "1.0");

        return header.str;
    }

    private GSignond.Dictionary? respond_with_stored_token (GSignond.Dictionary token) {
        var token_variant = token.get ("AccessToken");
        var token_secret = token.get ("TokenSecret");
        if (token_variant != null && token_secret != null) {
            var response = (GSignond.SessionData) new GSignond.Dictionary ();
            response.set ("AccessToken", token_variant);
            response.set ("TokenSecret", token_secret);
            var token_params = token.get("TokenParameters");
            if (token_params != null)
                response.set ("TokenParameters", token_params);

            return response;
        }

        return null;
    }

    private void reset () {
        oauth_request = null;
        token_cache = null;
    }

    [CCode (instance_pos = -1)]
    private void LoggerPrinter (Soup.Logger logger, Soup.LoggerLogLevel level, char direction, string data) {
        debug ("%c %s", direction, data);
    }
}
